import numpy as np
import time


class Timer(object):
    """
    Simple wrapper to time things easier.

    Examples
    --------
    The code below runs f()
    #
    # >>> from utility.timing import Timer
    # >>> with Timer():
    # >>>     f()

    and outputs the elapsed time.  Equivalent in matlab to
    tic; f(); toc;
    """

    def __init__(self, name=None):
        self.name = name

    def __enter__(self):
        self.tstart = time.time()

    def __exit__(self, type, value, traceback):
        if self.name:
            print(self.name)
        print('Elapsed: ', (time.time() - self.tstart), 's')


class iFTS_model(object):
    """
    This class models a simulated imaging FTS. The class can be used to
    produce simulated L0 interferograms given an input spectral radiance. The user can provide a yaml file
    containing the FTS specifications or the the class will use the default yaml file that is provided.

    To use this class all one has to do is instantiate it with a numpy array of wavelengths [nm] and radiance
    [photons sr^-1 nm^-1 cm^-2 s^-1] then call the functions set_input_spectrum(self, in_wavelengths, in_radiance) and
    then generate_L0_interferogram().



   """

    def __init__(self, specs):

        # Define properties.
        self.in_radiance = None
        self.in_wvnum = None
        self.specs = specs

        # Set the range of illuminated pixels
        self.x_pixel = np.arange(0, self.specs.DetNumPixX)
        self.y_pixel = np.arange(0, self.specs.DetNumPixY)

        # generate the angular grid
        self._generate_angular_grid()

        # generate the opd grid
        self._generate_opd_grid()

    def set_input_spectrum(self, in_wavelengths, in_radiance):
        """
         Sets the input spectrum (photons/s/cm**2/sr/nm) that is incident on the SHS detector by incorporating the detector properties
         and the transmission chanacteristics of the optical system

         input:
            in_wavelengths: wavelengths of the input spectrum
            in_radiance   : input radiance
        """
        # Set the wavenumbers of the input radiance
        self.in_wavelengths = in_wavelengths
        self.in_wvnum = self.__swap_wvnum_wvlen__(in_wavelengths)

        # Multiply by the number of vertical bins in an image and convert to (photons/s/cm**2/sr/cm-1)
        self.in_radiance = in_radiance * self.in_wavelengths ** 2 / 1e7

    def _generate_angular_grid(self):

        delta_alpha = (self.specs.interferometer_fov * 1e-3 / self.specs.DetNumPixY)

        self.alpha = np.zeros((len(self.x_pixel), len(self.y_pixel)))
        for x in self.x_pixel:
            for y in self.y_pixel:
                self.alpha[x, y] = delta_alpha * np.sqrt((x - self.specs.DetNumPixX / 2) ** 2 + (y - self.specs.DetNumPixY / 2) ** 2)

    def _generate_opd_grid(self):

        self.opd_o = np.linspace(-self.specs.max_opd, self.specs.max_opd, self.specs.sampling)
        self.angular_opd = np.zeros((len(self.x_pixel), len(self.y_pixel), len(self.opd_o)))

        # Off-axis optical path difference
        for opd_idx in range(len(self.opd_o)):
            self.angular_opd[:, :, opd_idx] = self.opd_o[opd_idx] * np.cos(self.alpha)

    def interferogram_sample(self, radiance, opd):

        """Output: interferogram image in units of [photons sr^-1 cm^-2 s^-1]"""

        iGM_modulated = 2 * self.specs.modulation_efficiency * radiance * self.specs.transmission_beam_1 * self.specs.transmission_beam_2 * (np.cos(2 * np.pi * self.in_wvnum * opd))

        return np.trapz(iGM_modulated, self.in_wvnum)

    def DC_sample(self, idx_x, idx_y):
        """Output: interferogram image in units of [photons sr^-1 cm^-2 s^-1]"""

        I1 = (self.specs.transmission_beam_1 ** 2) * np.trapz(
            self.in_radiance[idx_x, idx_y, :], self.in_wvnum)

        I2 = (self.specs.transmission_beam_2 ** 2) * np.trapz(
            self.in_radiance[idx_x, idx_y, :], self.in_wvnum)

        interferogram_DC = I1 + I2

        return interferogram_DC

    def __swap_wvnum_wvlen__(self, in_value):
        """
        This function swaps @in_value between wave number (cm-1) and wavelength (nm).
        """
        return 1e7 / in_value

    def image_pixel(self, c_x, c_y, opd_idx):

        image = self.interferogram_sample(self.in_radiance[c_x, c_y], self.angular_opd[c_x, c_y, opd_idx])

        return image

    def interferogram_image_sample(self, opd_idx):

        # Fill the image array at the chosen opd sample
        interferogram = np.zeros((self.specs.DetNumPixX, self.specs.DetNumPixY))
        interferogram_DC = np.zeros((self.specs.DetNumPixX, self.specs.DetNumPixY))

        for idx_y in range(self.specs.DetNumPixY):
            for idx_x in range(self.specs.DetNumPixX):

                interferogram[idx_x, idx_y] = self.image_pixel(idx_x, idx_y, opd_idx)
                interferogram_DC[idx_x, idx_y] = self.DC_sample(idx_x, idx_y)

        return interferogram + interferogram_DC

    def interferogram_image_scan(self):

        interferogram_image_scan = []
        for opd_idx in range(np.shape(self.angular_opd)[2]):
            with Timer():
                iGM = self.interferogram_image_sample(opd_idx)
            interferogram_image_scan.append(iGM)

        return interferogram_image_scan