from collections import OrderedDict
import numpy as np

from .noise import ADU_noise
from .noise import PoissonNoise
from .noise import RandomNoise
from .noise import ReadoutNoise
from .noise import DigitizationNoise
from .photons_to_electrons import electons_per_bin
from .noise import PRNU as prnu
from .adc import ADC as ADC


class level0_processing(object):

    def __init__(self, specs, digitization=True, integration_time=1, do_noise=True):

        self.specs = specs
        self.specs.integrationTime = integration_time
        self.integration_time = integration_time
        self.noise = do_noise
        self.digitization = digitization

        self._electronics = OrderedDict()
        self._electronics_noise = OrderedDict()

        self.__create_L0_processors__()
        self.__create_ADU_noise_processor__()

    def __create_L0_processors__(self):

        self.add_electronics_component(electons_per_bin(self.specs, self.integration_time, add_dark=True), 'dark')
        self.add_electronics_component(RandomNoise(self.specs, do_noise=self.noise), 'random')
        self.add_electronics_component(ReadoutNoise(self.specs, do_noise=self.noise), 'readout')
        self.add_electronics_component(ADC(digitization=self.digitization, min_value=0, max_value=self.specs.DetectorWell, bits=self.specs.DetectorBits, rescale=False), 'adc')

    def __create_ADU_noise_processor__(self):
        # Synthetic no-noise signal used to estimate the noise for ensemble tests
        self.add_electronics_noise_component(electons_per_bin(self.specs, self.integration_time, add_dark=True), 'dark')
        self.add_electronics_noise_component(ADC(digitization=self.digitization,
                                                 min_value=0, max_value=self.specs.DetectorWell,
                                                 bits=self.specs.DetectorBits,
                                                 rescale=False), 'adc')

    def add_electronics_component(self, processor, name):
        self._electronics[name] = processor

    def add_electronics_noise_component(self, processor, name):
        self._electronics_noise[name] = processor

    def integrate_pixel(self, model_value):

        # Take the level0_iGM and integrate over the pixel dimensions and solid angle to get the number of photons
        # output is photons per second

        model_value = model_value * self.specs.SolidAngle * self.specs.pupil_area

        return model_value

    def model_detected_signal(self, model_value):

        # Integrate over the projected pixel area
        model_value = self.integrate_pixel(model_value)

        process_array = []
        for name, processor in self._electronics.items():

            model_value = processor.process_signal(model_value)
            process_array.append(model_value)

        return model_value

    def model_ADU_noise(self, model_value):

        # Integrate over the projected pixel area
        model_value = self.integrate_pixel(model_value)

        process_array = []
        for name, processor in self._electronics.items():

            model_value = processor.process_signal(model_value)
            process_array.append(model_value)

        return ADU_noise(self.specs, self.integration_time, do_noise=True).noise_estimate_ADU(model_value)
