from typing import List
import numpy as np
import logging
import os
import os.path
import unittest
import time
from ifts_model.data.Specifications.iFTS_configuration import iFTS_specifications as specs
from ifts_model.optics.iFTS import iFTS_model as iFTS
from ifts_model.processing.level0 import level0_processing as level0
from ifts_model.data.Radiance.TestRadiance import TestRadiance as input_radiance

logging.getLogger().setLevel(logging.INFO)


# -----------------------------------------------------------------------------
#               class Timer(object):
# -----------------------------------------------------------------------------
class Timer(object):
    """
    Simple wrapper to time things easier.

    Examples
    --------
    The code below runs f()
    #
    # >>> from utility.timing import Timer
    # >>> with Timer():
    # >>>     f()

    and outputs the elapsed time.  Equivalent in matlab to
    tic; f(); toc;
    """

    def __init__(self, name=None):
        self.name = name

    def __enter__(self):
        self.tstart = time.time()

    def __exit__(self, type, value, traceback):
        if self.name:
            print(self.name)
        print('Elapsed: ', (time.time() - self.tstart), 's')


# -----------------------------------------------------------------------------
#               Test_iFTS_instrument_model
# -----------------------------------------------------------------------------
def Test_iFTS_instrument_model(specs, sampleindex: List[int]=None):

    # Set up the imaging FTS
    imaging_fts = iFTS(specs)

    # #Define the source
    # labsphere_wavelengths = LabSphere().wavelengths_nm
    # labsphere_radiance            = LabSphere().Radiance_photons

    # Define the filter

    directory_path, name = os.path.split(__file__)
    path = os.path.normpath(os.path.join(directory_path, '../data/Specifications/', 'filter_band_3_1.txt'))

    filter_file = np.loadtxt(path)
    filter_wavelengths = np.flip(filter_file[:, 0])
    filter_transmission = np.flip(filter_file[:, 1])

    # #interp to a higher wavelength grid
    # highres_wavelengths = np.arange(labsphere_wavelengths[0], labsphere_wavelengths[-1], 0.01)
    # radiance            = np.interp(highres_wavelengths, labsphere_wavelengths, labsphere_radiance)

    highres_wavelengths = input_radiance().wavelengths
    radiance = input_radiance().radiance
    
    # Pass through the filter
    filtered_radiance = np.interp(highres_wavelengths, filter_wavelengths, filter_transmission / 100) * radiance

    # set the input radiance
    imaging_fts.set_input_spectrum(highres_wavelengths, filtered_radiance * np.ones((specs.DetNumPixX, specs.DetNumPixY, len(highres_wavelengths))))

    # Generate a single interferogram image for a particular sample number

    samples = sampleindex if sampleindex is not None else range(len(imaging_fts.opd_o))
    # samples = len(imaging_fts.opd_o)
    image_scan = []
    for i in samples:
        # with Timer():
        single_image = imaging_fts.interferogram_image_sample(i)
        l0_processing = level0(specs=specs, integration_time=0.001)
        level0_iGM = l0_processing.model_detected_signal(single_image)
        image_scan.append(level0_iGM)
            
    return image_scan


# -----------------------------------------------------------------------------
#               def process_all_pixels():
# -----------------------------------------------------------------------------
def process_all_pixels():

    specs_file = r'iFTS_specifications.yaml'
    directory_path, name = os.path.split(__file__)
    path = os.path.normpath(os.path.join(directory_path, '../data/Specifications', specs_file))

    iFTS_config = specs(filename=path, channel='O2')
    output = Test_iFTS_instrument_model(iFTS_config)
    return output


# -----------------------------------------------------------------------------
#               class ifts_testmodel(unittest.TestCase):
# -----------------------------------------------------------------------------
class test_ifts_model(unittest.TestCase):

    def test_instrumentmodel(self):
        '''
        very similar to Jeff's original except this only calls three pixels and takes about 3 seconds to run rather than a half hour-hour
        '''
        specs_file = r'iFTS_specifications.yaml'
        directory_path, name = os.path.split(__file__)
        path = os.path.normpath(os.path.join(directory_path, '../data/Specifications', specs_file))
        iFTS_config = specs(filename=path, channel='O2')
        samples = [84, 1596, 3108]
        output = Test_iFTS_instrument_model(iFTS_config, sampleindex=samples)
        print('Done')


if __name__ == "__main__":
    tests = test_ifts_model()
    unittest.main()
